@@ -35,8 +35,8 @@ SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 void intel_batch_init(ScrnInfoPtr scrn);
 void intel_batch_teardown(ScrnInfoPtr scrn);
-void intel_batch_pipelined_flush(ScrnInfoPtr scrn);
-void intel_batch_flush(ScrnInfoPtr scrn, Bool flush);
+void intel_batch_emit_flush(ScrnInfoPtr scrn);
+void intel_batch_submit(ScrnInfoPtr scrn);
 void intel_batch_wait_last(ScrnInfoPtr scrn);
 
 static inline int intel_batch_space(intel_screen_private *intel)
@@ -49,7 +49,7 @@ intel_batch_require_space(ScrnInfoPtr scrn, intel_screen_private *intel, GLuint
 {
 	assert(sz < intel->batch_bo->size - 8);
 	if (intel_batch_space(intel) < sz)
-		intel_batch_flush(scrn, FALSE);
+		intel_batch_submit(scrn);
 }
 
 static inline void intel_batch_start_atomic(ScrnInfoPtr scrn, unsigned int sz)
@@ -75,11 +75,25 @@ static inline void intel_batch_end_atomic(ScrnInfoPtr scrn)
 static inline void intel_batch_emit_dword(intel_screen_private *intel, uint32_t dword)
 {
 	assert(intel->batch_ptr != NULL);
-	assert(intel_batch_space(intel) >= 4);
+	assert(intel->batch_emitting);
 	*(uint32_t *) (intel->batch_ptr + intel->batch_used) = dword;
 	intel->batch_used += 4;
 }
 
+static inline void intel_batch_align(intel_screen_private *intel, uint32_t align)
+{
+	uint32_t delta;
+
+	assert(intel->batch_ptr != NULL);
+	assert(align);
+
+	if ((delta = intel->batch_used & (align - 1))) {
+		delta = align - delta;
+		memset (intel->batch_ptr + intel->batch_used, 0, delta);
+		intel->batch_used += delta;
+	}
+}
+
 static inline void
 intel_batch_emit_reloc(intel_screen_private *intel,
 		       dri_bo * bo,
@@ -132,6 +146,7 @@ intel_batch_emit_reloc_pixmap(intel_screen_private *intel, PixmapPtr pixmap,
 			       delta);
 }
 
+#define ALIGN_BATCH(align) intel_batch_align(intel, align);
 #define OUT_BATCH(dword) intel_batch_emit_dword(intel, dword)
 
 #define OUT_RELOC(bo, read_domains, write_domains, delta) \
@@ -193,7 +208,7 @@ do {									\
 	if ((intel->batch_emitting > 8) &&				\
 	    (I810_DEBUG & DEBUG_ALWAYS_SYNC)) {				\
 		/* Note: not actually syncing, just flushing each batch. */ \
-		intel_batch_flush(scrn, FALSE);			\
+		intel_batch_submit(scrn);			\
 	}								\
 	intel->batch_emitting = 0;					\
 } while (0)
